// ----------------------------------------------------------------------------
//
// Matrix multiplication (thread pool version).
// (c) Wolfgang Schreiner, 2005.
//
// ----------------------------------------------------------------------------
class MatMultThreads2
{
  // matrix dimension
  final static int N = 1024; 

  // number of threads
  static int T;
 
  // the threads
  static Thread manager;
  static MultThread[] worker;

  // number of completed workers
  static int done;

  // shared data used by threads
  static float[][] A0;
  static float[][] B0;
  static float[][] C0;

  static class MultThread extends Thread
  {
    int begin;    // index of first row
    int end;      // index of last row + 1
    boolean dead; // signal to die

    // -------------------------------------------------------------------------
    // 
    // Create thread in charge of filling C[begin..end-1]
    //
    // -------------------------------------------------------------------------
    public MultThread(int begin, int end)
    {
      this.begin = begin;
      this.end = end;
      this.dead = false;
    }

    // ----------------------------------------------------------------------------
    // 
    // multiply A by B giving C
    // 
    // ----------------------------------------------------------------------------
    synchronized public void run()
    {
      while (true)
      {
        // awake manager
        synchronized(manager) 
        { 
          done++;
          if (done == T) manager.notify();
        }

        // wait until awaked
        try
        { wait(); }
        catch(InterruptedException e)
        { }
        if (dead) return;

        for (int i = begin; i < end; i++) 
        {
          for (int j = 0; j < N; j++) 
          {
	    float m = 0;
	    for (int k = 0; k < N; k++) 
	      m = m + A0[i][k]*B0[k][j];
	    C0[i][j] = m;
          }
        }
      }
    }
  }

  // ----------------------------------------------------------------------------
  // 
  // multiply A by B giving C
  // 
  // ----------------------------------------------------------------------------
  static void multiply(float[][] A, float[][] B, float[][] C) 
  {
    A0 = A;
    B0 = B;
    C0 = C;

    // notify tasks
    done = 0;
    for (int i = 0; i < T; i++) 
      synchronized(worker[i]) { worker[i].notify(); }
 
    // wait until awaked
    try
    { synchronized(manager) { manager.wait(); } }
    catch(InterruptedException e)
    { }
  }

  // ----------------------------------------------------------------------------
  //
  // main program
  //
  // ----------------------------------------------------------------------------
  public static void main(String[] args) 
  {
    if (args.length != 1)
    {
      System.out.println("Usage: MatMultThreads <number of threads>");
      return;
    }
    try
    {
      T = Integer.parseInt(args[0]);
    }
    catch(NumberFormatException e)
    {
      System.out.println("Invalid argument (number expected): " + args[0]);
      return;
    } 

    manager = Thread.currentThread();
    worker = new MultThread[T];
    int n = N/T;

    synchronized(manager)
    {
      for (int i = 0; i < T; i++) 
      {
	if (i+1 < T)
	  worker[i] = new MultThread(i*n, i*n+n);
	else
	  worker[i] = new MultThread(i*n, N);
	worker[i].start();
      }
      done = 0;
      try { manager.wait(); } catch(InterruptedException e) { }
    }  

    for (int i = 0; i < 1; i ++)
    {
      float[][] A = new float[N][N];
      float[][] B = new float[N][N];
      float[][] C = new float[N][N];
      multiply(A, B, C);
    }

    for (int i = 0; i < T; i++)
    {
      synchronized(worker[i]) 
      { 
        worker[i].dead = true; 
        worker[i].notify(); 
      }
      try
      { worker[i].join(); }
      catch (InterruptedException e)
      { }
    }
  }
}

